package cn.spatiotemporal.core.async.agent;


import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.lang.reflect.Type;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import cn.spatiotemporal.core.async.agent.InvocationHandlerFactory.MethodHandler;
import cn.spatiotemporal.core.async.agent.reflect.ReflectUtils;
import cn.spatiotemporal.core.async.agent.reflect.Target;
import cn.spatiotemporal.core.async.agent.reflect.Target.EmptyTarget;
import cn.spatiotemporal.core.async.agent.reflect.Types;



public class AsyncAgent {

	private final InvocationHandlerFactory factory;
	
	private String topic;
	
	public AsyncAgent(InvocationHandlerFactory factory, String topic) {
		this.factory = factory;
		this.topic = topic;
	}
	
	public static AsyncAgent builder(String topic) {
		Builder builder = new Builder();
		return builder.build(topic);
	}

	public static String configKey(Class targetType, Method method) {
		StringBuilder builder = new StringBuilder();
		builder.append(targetType.getSimpleName());
		builder.append('#').append(method.getName()).append('(');
		for (Type param : method.getGenericParameterTypes()) {
			param = Types.resolve(targetType, targetType, param);
			builder.append(Types.getRawType(param).getSimpleName()).append(',');
		}
		if (method.getParameterTypes().length > 0) {
			builder.deleteCharAt(builder.length() - 1);
		}
		return builder.append(')').toString();
	}

	public <T> T newInstance(Target<T> target) {
		Map<Method, MethodHandler> methodToHandler = new LinkedHashMap<Method, MethodHandler>();
		List<DefaultMethodHandler> defaultMethodHandlers = new LinkedList<DefaultMethodHandler>();
		List<AsyncMethodHandler> asyncMethodHandlers = new LinkedList<AsyncMethodHandler>();

		for (Method method : target.type().getMethods()) {
			if (method.getDeclaringClass() == Object.class) {
				continue;
			} else if (ReflectUtils.isDefault(method)) {
				DefaultMethodHandler handler = new DefaultMethodHandler(method);
				defaultMethodHandlers.add(handler);
				methodToHandler.put(method, handler);
			} else {
				AsyncMethodHandler handler = new AsyncMethodHandler(target, method);
				asyncMethodHandlers.add(handler);
				methodToHandler.put(method, handler);
			}
		}
		InvocationHandler handler = factory.create(target, methodToHandler);
		T proxy = (T) Proxy.newProxyInstance(target.type().getClassLoader(), new Class<?>[] { target.type() }, handler);

		for (DefaultMethodHandler defaultMethodHandler : defaultMethodHandlers) {
			defaultMethodHandler.bindTo(proxy);
		}
		return proxy;
	}

	static class AsyncInvocationHandler implements InvocationHandler {

		private final Target target;
		private final Map<Method, MethodHandler> dispatch;

		AsyncInvocationHandler(Target target, Map<Method, MethodHandler> dispatch) {
			this.target = target;
			this.dispatch = ReflectUtils.checkNotNull(dispatch, "dispatch for %s", target);
		}

		@Override
		public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
			if ("equals".equals(method.getName())) {
				try {
					Object otherHandler = args.length > 0 && args[0] != null ? Proxy.getInvocationHandler(args[0])
							: null;
					return equals(otherHandler);
				} catch (IllegalArgumentException e) {
					return false;
				}
			} else if ("hashCode".equals(method.getName())) {
				return hashCode();
			} else if ("toString".equals(method.getName())) {
				return toString();
			}
			return dispatch.get(method).invoke(args);
		}

		@Override
		public boolean equals(Object obj) {
			if (obj instanceof AsyncInvocationHandler) {
				AsyncInvocationHandler other = (AsyncInvocationHandler) obj;
				return target.equals(other.target);
			}
			return false;
		}

		@Override
		public int hashCode() {
			return target.hashCode();
		}

		@Override
		public String toString() {
			return target.toString();
		}
	}
	
	public <T> T build(Class<T> type, String name, Map<String, String> attrs) {
		EmptyTarget<T> target = EmptyTarget.create(type, name);
		return newInstance(target);
	}

    public static class Builder {

		// TODO 获取MQ配置
		//private Producer producer = ConnectionFactory.openConnect().getProducer();
	    private InvocationHandlerFactory invocationHandlerFactory =
	            new InvocationHandlerFactory.Default();
	    
		public AsyncAgent build(String topic) {
			return new AsyncAgent(invocationHandlerFactory, topic);
		}
	}
}
