package cn.spatiotemporal.core.async.client;

import java.lang.reflect.ParameterizedType;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import java.util.function.Consumer;

import org.springframework.amqp.rabbit.annotation.RabbitListener;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Component;

import cn.spatiotemporal.core.async.AsyncConstants;
import cn.spatiotemporal.core.async.AsyncParamMessage;
import cn.spatiotemporal.core.async.AsyncReturnMessage;
import cn.spatiotemporal.core.async.spring.EnableAsyncClient;
import cn.spatiotemporal.core.utils.FastJsonUtils;

import net.jodah.typetools.TypeResolver;

/**
 * 异步调用的客户端处理
 * @author marquis
 *
 */
@Component
@ConditionalOnBean(annotation = EnableAsyncClient.class)
public class AsyncClient {
	
	@Autowired
	private	String compInstanceId;
	
	@Autowired
	private RabbitTemplate template;
	
	/**
	 * 成功回调实例储存Map
	 */
	private static Map<String, Consumer<?>> onSuccussCallbacks = new HashMap<String, Consumer<?>>();
	
	/**
	 * 失败回调实例储存Map
	 */
	private static Map<String, Consumer<String>> onFailCallbacks = new HashMap<String, Consumer<String>>();
	
	/**
	 * 回调返回值的类型
	 */
	private Class resultType;
	
	/**
	 * 异步调用
	 * @param compName 远程模块
	 * @param className 对象类名称
	 * @param methodName 方法名称
	 * @param params 参数
	 */
	public AsyncResult call(String compName, String className, String methodName, Object[] params) {
		AsyncResult result = new AsyncResult(); 
		
		//随机生成异步调用实例ID
		String instanceId = String.valueOf(UUID.randomUUID().getMostSignificantBits());
		result.setInstanceId(instanceId);
		
		//发送异步调用消息
		AsyncParamMessage param = new AsyncParamMessage();
		param.setCid(compInstanceId);
		param.setInstanceId(instanceId);
		param.setTarget(className);
		param.setMethod(methodName);
		param.setParams(FastJsonUtils.object2Json(params));
		template.convertAndSend(AsyncConstants.ASYNC_CALL_MQ + compName,
				param);
		
		return result;
	}
	
	/**
	 * 异步调用
	 * @param compName 远程模块
	 * @param className 对象类名称
	 * @param methodName 方法名称
	 * @param params 参数
	 */
	public <T> AsyncResult call(String compName, String className, String methodName, Object[] params, Class<T> clazz, Consumer<T> consumer) {
		//注册成功回调实例
		return call(compName, className, methodName, params).onSuccuss(clazz, consumer);
	}

	/**
	 * 异步调用结果处理
	 * @param returnMsg
	 */
	@RabbitListener(queues = AsyncConstants.ASYNC_RETURN_MQ + "#{@compInstanceId}")
	public void resultHandle(AsyncReturnMessage returnMsg) {
		try {
			//如果异步调用执行有异常，则调用异常回调处理，否则抛出异常
			if (returnMsg.isHasException()) {
				//获取回调实例
				Consumer<String> onFail = (Consumer<String>) onFailCallbacks.get(returnMsg.getInstanceId());
				if (onFail != null) {
					onFail.accept(returnMsg.getErrorMessage());
				} else {
					//TODO throw异常 无法结束
					System.out.println(returnMsg.getErrorMessage());
					//throw new AsyncProcessException(returnMsg.getErrorMessage());
				}
			} else {
				//获取回调实例
				Consumer onSuccuss = onSuccussCallbacks.get(returnMsg.getInstanceId());
				if (onSuccuss != null) {
					Class<?> clazz = TypeResolver.resolveRawArguments(Consumer.class, onSuccuss.getClass())[0];
					onSuccuss.accept(FastJsonUtils.json2Object(returnMsg.getReturnValue(), clazz));
				}
			}
		} catch (Exception e) {
			// TODO 记录异步调用异常，记录日志
			e.printStackTrace();
		}
	}
	
	/**
	 * 注册正常回调处理
	 * @param instanceId
	 * @param consumer
	 */
	protected static <T> void registOnSuccussCallback(String instanceId, Class<T> clazz, Consumer<T> consumer) {
		if (consumer != null) {
			onSuccussCallbacks.put(instanceId, consumer);
		}
	}
	
	/**
	 * 注册异常回调处理
	 * @param instanceId
	 * @param consumer
	 */
	protected static void registOnFailCallback(String instanceId, Consumer<String> consumer) {
		if (consumer != null) {
			onFailCallbacks.put(instanceId, consumer);
		}
	}
	
	private Class getResultType() {
		if (resultType == null) {
			resultType = (Class) ((ParameterizedType)this.getClass().getGenericSuperclass())
					.getActualTypeArguments()[0];    //得到泛型类型		
		}
		return resultType;
	}
}
