package cn.spatiotemporal.core.async.server;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.util.List;
import java.util.Map;

import org.springframework.amqp.rabbit.annotation.RabbitListener;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.stereotype.Component;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import cn.spatiotemporal.core.async.AsyncConstants;
import cn.spatiotemporal.core.async.AsyncParamMessage;
import cn.spatiotemporal.core.async.AsyncReturnMessage;
import cn.spatiotemporal.core.async.exception.AsyncProcessException;
import cn.spatiotemporal.core.async.spring.EnableAsyncServer;
import cn.spatiotemporal.core.utils.FastJsonUtils;
import cn.spatiotemporal.core.utils.SpringBeanUtils;

@Component
@ConditionalOnBean(annotation = EnableAsyncServer.class)
public class AsyncDispatcher {
	
	@Autowired
	private String compInstanceId;
	
	@Autowired
	private RabbitTemplate template;
	
	/**
	 * 异步调用调度处理
	 * @param param
	 */
	@RabbitListener(queues = AsyncConstants.ASYNC_CALL_MQ + "#{@compId}")
	public void dispath(AsyncParamMessage param){
		AsyncReturnMessage returnMsg = new AsyncReturnMessage();
		returnMsg.setEid(compInstanceId);
		returnMsg.setInstanceId(param.getInstanceId());
		returnMsg.setHasException(false);
		
		//查找调用对象及方法，然后执行
		try {
			Object target = SpringBeanUtils.getBean(param.getTarget());
			if (target == null) {
				throw new AsyncProcessException("无法找到服务对象["+param.getTarget()+"]");
			}
			String paramsJson = param.getParams();
			JSONArray jarray = JSON.parseArray(paramsJson);
			Method method = findMethod(target.getClass(), param.getMethod(), jarray.size());
			if (method == null) {
				throw new AsyncProcessException("无法匹配到合适的方法["+param.getMethod()+"]");
			}
			Object[] params = json2Params(jarray, method.getParameterTypes());
			Object returnValue = method.invoke(target, params);
			returnMsg.setReturnValue(FastJsonUtils.object2Json(returnValue));
		} catch (Exception e) {
			returnMsg.setHasException(true);
			returnMsg.setErrorMessage(e.getMessage());
		}
		
		//发送返回消息
		template.convertAndSend(AsyncConstants.ASYNC_RETURN_MQ + param.getCid(),
				returnMsg);
	}
	
	/**
	 * 查找对象方法
	 * @param clazz
	 * @param name
	 * @param paramSize
	 * @return
	 */
	private Method findMethod(Class<?> clazz, String name, int paramSize) {
		for (Method method : clazz.getMethods()) {
			if (name.equals(method.getName()) && method.getParameterCount() == paramSize) {
				return method;
			}
		}
		return null;
	}
	
	private Object[] json2Params(JSONArray jarray, Class<?>[] clazzes) {
		Object[] params = new Object[jarray.size()];
		
		for (int i = 0; i < jarray.size(); i++) {
			Class clazz =  clazzes[i];
			if (clazz.isArray()) {
				params[i] = FastJsonUtils.json2Array(jarray.getJSONArray(i), clazz.getComponentType());
			} else if (clazz.isAssignableFrom(List.class)) {
				Class type = (Class) ((ParameterizedType)clazz.getGenericSuperclass())
						.getActualTypeArguments()[0];    //得到泛型类型
				params[i] = FastJsonUtils.json2List(jarray.getJSONArray(i), type);
			} else if (clazz.isAssignableFrom(Map.class)) {
				Class keyType = (Class) ((ParameterizedType)clazz.getGenericSuperclass())
						.getActualTypeArguments()[0];    //得到key的泛型类型
				Class valueType = (Class) ((ParameterizedType)clazz.getGenericSuperclass())
						.getActualTypeArguments()[1];    //得到value的泛型类型
				params[i] = FastJsonUtils.json2Map(jarray.getJSONObject(i), keyType, valueType);
			} else if (FastJsonUtils.isSimple(clazz)) {
				// 基本类型
				params[i] = jarray.getObject(i, clazz);
			} else {
				// 一般Bean
				params[i] = FastJsonUtils.json2Bean(jarray.getJSONObject(i), clazz);
			}
		}
		return params;
	}
}
