package cn.spatiotemporal.core.ftp;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPFile;
import org.apache.commons.pool2.ObjectPool;
import org.springframework.stereotype.Component;
import org.springframework.util.Assert;

import cn.spatiotemporal.core.config.ftp.FtpProperties;

import lombok.Getter;
import lombok.Setter;
import lombok.extern.slf4j.Slf4j;

@Slf4j
@Getter
@Setter
public class FtpTemplate implements FtpDao {


    private final FtpProperties ftpProperties;
 
    //连接池初始化标志
    private boolean hasInit = false;
    //连接池
    private ObjectPool<FTPClient> ftpClientPool;
    
    private FTPClient currentClient = null;
 
    public FtpTemplate(FtpProperties ftpProperties) {
        this.ftpProperties = ftpProperties;
    }
 
    /**
     * 上传文件
     *
     * @param path           ftp服务器保存地址
     * @param fileName       上传到ftp的文件名
     * @param originFileName 等待上传的文件名（绝对地址或路径）
     */
    @Override
    public boolean uploadFile(String path, String fileName, String originFileName) {
        boolean flag = false;
        try {
            InputStream inputStream  = new FileInputStream(new File(originFileName));
            flag=uploadFile( path,  fileName,  inputStream);
        } catch (Exception e) {
            //log.error("上传文件出错！", (Object) e.getStackTrace());
            e.printStackTrace();
        }
        return flag;
    }
 
    /**
     * 上传文件
     *
     * @param path        ftp服务器保存地址
     * @param fileName    上传到ftp的文件名
     * @param inputStream 文件流
     */
    @Override
    public boolean uploadFile(String path, String fileName, InputStream inputStream) throws IOException {
        boolean flag = false;
        currentClient = getFtpClient();
        try {
            currentClient.setFileType(FTPClient.BINARY_FILE_TYPE);
            createDirectory(path);
            currentClient.makeDirectory(path);
            currentClient.changeWorkingDirectory(path);
            currentClient.storeFile(fileName, inputStream);
            flag = true;
        } catch (Exception e) {
            //log.error("上传文件出错！:{}", e);
            e.printStackTrace();
        } finally {
            if(null!=inputStream) inputStream.close();
            releaseFtpClient(currentClient);
            currentClient = null;
        }
        return flag;
    }
 
    /**
     * 下载文件
     *
     * @param path      ftp服务器文件路径
     * @param fileName  文件名称
     * @param localPath 下载后的路径
     */
    @Override
    public boolean downloadFile(String path, String fileName, String localPath) {
        boolean flag = false;
        OutputStream outputStream = null;
        currentClient = getFtpClient();
        try {
            currentClient.changeWorkingDirectory(path);
            FTPFile[] files = currentClient.listFiles();
            for (FTPFile file : files) {
                if (fileName.equalsIgnoreCase(file.getName())) {
                    File localFile = new File(localPath + "/" + file.getName());
                    outputStream = new FileOutputStream(localFile);
                    currentClient.retrieveFile(file.getName(), outputStream);
                    outputStream.close();
                }
            }
            flag = true;
        } catch (IOException e) {
            log.error("下载文件出错！", (Object) e.getStackTrace());
        } finally {
            releaseFtpClient(currentClient);
            if (null != outputStream) {
                try {
                    outputStream.close();
                } catch (IOException e) {
                    log.error("关闭输出流出错！", (Object) e.getStackTrace());
                }
            }
            currentClient = null;
        }
        return flag;
    }
 
    /**
     * 删除文件
     *
     * @param path     ftp文件路径
     * @param fileName 文件名
     */
    @Override
    public boolean deleteFile(String path, String fileName) {
        boolean flag = false;
        currentClient = getFtpClient();
        try {
            currentClient.changeWorkingDirectory(path);
            currentClient.dele(fileName);
            currentClient.logout();
            flag = true;
        } catch (IOException e) {
            log.error("删除文件出错！", (Object) e.getStackTrace());
        } finally {
            releaseFtpClient(currentClient);
            currentClient = null;
        }
        return flag;
    }
 
    /**
     * 创建多层目录，如果ftp服务器已存在该目录，则不创建，如果没有，则创建
     *
     * @param remote    创建的目录
     * @param currentClient
     */
    @Override
    public boolean createDirectory(String remote) {
        String directory = remote + "/";
        boolean flag = true;
        if (currentClient == null) {
        	currentClient = getFtpClient();
        }
        try {
        	//如果远程目录不存在，则递归创建远程目录
			if (!directory.equalsIgnoreCase("/") && !changeWorkingDirectory(directory)) {
			    int start = 0;
			    int end = 0;
			    if (directory.startsWith("/")) {
			        start = 1;
			    }
			    end = directory.indexOf("/", start);
			    String path = "";
			    String paths = "";
			    do {
			        String subDirectory = new String(remote.substring(start, end).getBytes(ftpProperties.getEncoding()), FtpConstants.DEFAULT_FTP_PATH_ENCODING);
			        path = path + "/" + subDirectory;
			        if (!existFile(path)) {
			            if (makeDirectory(subDirectory)) {
			                changeWorkingDirectory(subDirectory);
			            } else {
			                log.warn("创建目录[" + subDirectory + "]失败");
			                changeWorkingDirectory(subDirectory);
			                flag = false;
			            }
			        } else {
			            changeWorkingDirectory(subDirectory);
			        }
			        paths = paths + "/" + subDirectory;
			        start = end + 1;
			        end = directory.indexOf("/", start);
			    } while (end <= start);
			}
		} catch (IOException e) {
			log.error("创建目录["+remote+"]失败！", (Object) e.getStackTrace());
			flag = false;
	    }
        return flag;
    }
 
    /**
     * 判断ftp服务器的路径或文件是否存在
     *
     * @param path
     * @param currentClient
     */
    @Override
    public boolean existFile(String path) {
        boolean flag = false;
        if (currentClient == null) {
        	currentClient = getFtpClient();
        }
        try {
			FTPFile[] files = currentClient.listFiles(path);
			if (files.length > 0) {
			    flag = true;
			}
		} catch (IOException e) {
			log.error("判断目录["+path+"]是否存在失败！", (Object) e.getStackTrace());
			flag = false;
		}
        return flag;
    }
 
    /**
     * 创建目录
     *
     * @param directory
     * @param currentClient
     */
    @Override
    public boolean makeDirectory(String directory) {
        boolean flag = true;
        if (currentClient == null) {
        	currentClient = getFtpClient();
        }
        try {
            flag = currentClient.makeDirectory(directory);
            if (flag) {
                log.info("创建文件夹：" + directory);
            }
        } catch (IOException e) {
            log.error("创建文件夹" + directory + "失败！", (Object) e.getStackTrace());
            flag = false;
        }
        return flag;
    }
 
    /**
     * 切换目录
     *
     * @param directory 要切换的目录
     * @param currentClient ftp客户端
     */
    public boolean changeWorkingDirectory(String directory) {
        boolean flag = true;
        if (currentClient == null) {
        	currentClient = getFtpClient();
        }
        try {
            flag = currentClient.changeWorkingDirectory(directory);
            if (flag) {
                log.info("进入文件夹：" + directory);
            }
        } catch (IOException e) {
            log.error("进入文件夹："+directory+"错误！", (Object) e.getStackTrace());
            flag = false;
        }
        return flag;
    }
 
    /**
     * 按行读取FTP文件
     *
     * @param remoteFilePath ftp路径
     */
    public List<String> readFileByLine(String remoteFilePath) throws IOException {
        currentClient = getFtpClient();
        try (InputStream inputStream = currentClient.retrieveFileStream(encodingPath(remoteFilePath));
             BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream))) {
            return reader.lines()
                    .map(line -> StringUtils.trimToEmpty(line))
                    .filter(line -> StringUtils.isNotEmpty(line))
                    .collect(Collectors.toList());
        } finally {
            currentClient.completePendingCommand();
            releaseFtpClient(currentClient);
            currentClient = null;
        }
    }
 
    /**
     * 获取指定路径下的ftp文件
     *
     * @param remotePath 指定路径
     */
    public FTPFile[] retrieveFtpFiles(String remotePath) {
        currentClient = getFtpClient();
        try {
            return currentClient.listFiles(encodingPath(remotePath + "/"), file -> file != null && file.getSize() > 0);
        } catch (IOException e) {
        	log.error("获取路径["+remotePath+"]下的文件列表失败！", (Object) e.getStackTrace());
        	return null;
        } finally {
            releaseFtpClient(currentClient);
            currentClient = null;
        }
    }
 
    /**
     * 获取指定ftp路径下的文件名称
     *
     * @param remotePath 指定ftp路径
     */
    @Override
    public List<String> retrieveFileNames(String remotePath) {
        FTPFile[] files = retrieveFtpFiles(remotePath);
        if (ArrayUtils.isEmpty(files)) {
            return new ArrayList<>();
        }
        return Arrays.stream(files).filter(Objects::nonNull).map(FTPFile::getName).collect(Collectors.toList());
    }
 
    /**
     * 获取编码后的文件路径
     */
    private String encodingPath(String path) throws UnsupportedEncodingException {
        //在FTP协议中，规定文件名编码格式为ISO-8859-1，所以目录名或文件名需要转码
        return new String(path.replaceAll("//", "/").getBytes(ftpProperties.getEncoding()), FtpConstants.DEFAULT_FTP_PATH_ENCODING);
    }
 
    /**
     * 获取ftp客户端
     */
    private FTPClient getFtpClient() {
        checkFtpClientPoolAvailable();
        FTPClient ftpClient = null;
        Exception exception = null;
        //获取连接，做多尝试n次
        try {
            for (int i = 0; i < ftpProperties.getRetryCount(); i++) {
                ftpClient = ftpClientPool.borrowObject();
                ftpClient.enterLocalPassiveMode();//设置为被动模式
                ftpClient.changeWorkingDirectory("/");
                break;
            }
        } catch (Exception e) {
            log.error("无法在连接池中获取ftp客户端！", (Object) e.getStackTrace());
            exception = e;
        }
        if (null == ftpClient) {
            throw new RuntimeException("无法在连接池中获取ftp客户端", exception);
        }
        return ftpClient;
    }
 
    /**
     * 释放ftp客户端
     *
     * @param ftpClient
     */
    private void releaseFtpClient(FTPClient ftpClient) {
        if (null != ftpClient) {
            try {
                //从ftp连接池中移除ftp客户端
                ftpClientPool.returnObject(ftpClient);
            } catch (Exception e) {
                try {
                    //判断客户端是否可用
                    if (ftpClient.isAvailable()) {
                        //销毁连接
                        ftpClient.disconnect();
                    }
                } catch (IOException ex) {
                    log.error("销毁ftp连接失败！", (Object) e.getStackTrace());
                }
                log.error("从ftp连接池移除ftp客户端失败！", (Object) e.getStackTrace());
            }
        }
    }
 
    /**
     * 检查ftp连接池是否可用
     */
    private void checkFtpClientPoolAvailable() {
        Assert.state(hasInit, "ftp未启用或连接失败！");
    }
}
