package cn.spatiotemporal.core.redis;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;

import javax.annotation.Resource;

import org.springframework.data.redis.connection.RedisConnection;
import org.springframework.data.redis.core.Cursor;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.data.redis.core.ScanOptions;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;

import com.alibaba.fastjson.JSONObject;

/**
 * Redis访问操作类
 * 
 * @author marquis
 *
 */
@Component
public final class RedisDao {

	@Resource
	private RedisTemplate redisTemplate;
	
	public RedisTemplate template() {
		return redisTemplate;
	}

	public Set<String> keys(String keys) {
		try {
			return redisTemplate.keys(keys);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 指定缓存失效时间
	 * 
	 * @param key  键
	 * @param time 时间(秒)
	 * @return
	 */
	public boolean expire(String key, long time) {
		try {
			if (time > 0) {
				redisTemplate.expire(key, time, TimeUnit.SECONDS);
			}
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 指定缓存失效时间
	 * 
	 * @param key  键
	 * @param time 时间(日期)
	 * @return
	 */
	public boolean expireAt(String key, Date time) {
		try {
			redisTemplate.expireAt(key, time);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 根据key 获取过期时间
	 * 
	 * @param key 键 不能为null
	 * @return 时间(秒) 返回0代表为永久有效
	 */
	public long getExpire(String key) {
		return redisTemplate.getExpire(key, TimeUnit.SECONDS);
	}

	/**
	 * 判断key是否存在
	 * 
	 * @param key 键
	 * @return true 存在 false不存在
	 */
	public boolean hasKey(String key) {
		try {
			return redisTemplate.hasKey(key);
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 删除缓存
	 * 
	 * @param key 可以传一个值 或多个
	 */
	@SuppressWarnings("unchecked")
	public void del(String... key) {
		if (key != null && key.length > 0) {
			if (key.length == 1) {
				redisTemplate.delete(key[0]);
			} else {
				redisTemplate.delete(CollectionUtils.arrayToList(key));
			}
		}
	}

	/**
	 * 普通缓存获取
	 * 由于redisDao添加了自动序列化和反序列化工具(FastJson)，如果是对象的话get出来的会成为JSONObject对象，无法直接cast。\n
	 * 推荐使用getString()或getBean()方法来获取结果。
	 * 
	 * @param key 键
	 * @return 值
	 */
	@Deprecated
	public Object get(String key) {
		return key == null ? null : redisTemplate.opsForValue().get(key);
	}

	/**
	 * 普通缓存获取字符串
	 * 
	 * @param key 键
	 * @return 值
	 */
	public String getString(String key) {
		return key == null ? null : (String) redisTemplate.opsForValue().get(key);
	}

	/**
	 * 普通缓存获取对象
	 * 
	 * @param     <T>
	 * @param key 键
	 * @return 值
	 */
	public <T> T getBean(String key, Class<T> clazz) {
		JSONObject obj = (JSONObject) redisTemplate.opsForValue().get(key);
		// 反序列化
		return obj == null ? null : obj.toJavaObject(clazz);
	}

	/**
	 * 普通缓存放入
	 * 
	 * @param key   键
	 * @param value 值
	 * @return true成功 false失败
	 */
	public boolean put(String key, Object value) {
		try {
			redisTemplate.opsForValue().set(key, value);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 普通缓存放入并设置时间
	 * 
	 * @param key   键
	 * @param value 值
	 * @param time  时间(秒) time要大于0 如果time小于等于0 将设置无限期
	 * @return true成功 false 失败
	 */
	public boolean put(String key, Object value, long time) {
		try {
			if (time > 0) {
				redisTemplate.opsForValue().set(key, value, time, TimeUnit.SECONDS);
			} else {
				put(key, value);
			}
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 递增
	 * 
	 * @param key   键
	 * @param delta 要增加几(大于0)
	 * @return
	 */
	public long incr(String key, long delta) {
		if (delta < 0) {
			throw new RuntimeException("递增因子必须大于0");
		}
		return redisTemplate.opsForValue().increment(key, delta);
	}

	/**
	 * 递减
	 * 
	 * @param key   键
	 * @param delta 要减少几(小于0)
	 * @return
	 */
	public long decr(String key, long delta) {
		if (delta < 0) {
			throw new RuntimeException("递减因子必须大于0");
		}
		return redisTemplate.opsForValue().increment(key, -delta);
	}

	// ===============================hash=================================
	/**
	 * 获取HashMap中的值
	 * 
	 * @param key  键 不能为null
	 * @param item 项 不能为null
	 * @return 值
	 */
	public Object getHashValue(String key, String item) {
		return key == null || item == null ? null : redisTemplate.opsForHash().get(key, item);
	}

	/**
	 * 获取HashMap中的字符串
	 * 
	 * @param key  键 不能为null
	 * @param item 项 不能为null
	 * @return 值
	 */
	public String getHashString(String key, String item) {
		return key == null || item == null ? null : (String)redisTemplate.opsForHash().get(key, item);
	}

	/**
	 * 获取HashMap中的对象
	 * 
	 * @param key  键 不能为null
	 * @param item 项 不能为null
	 * @return 值
	 */
	public <T> T getHashBean(String key, String item, Class<T> clazz) {
		JSONObject obj = (JSONObject) redisTemplate.opsForHash().get(key, item);
		// 反序列化
		return obj == null ? null : obj.toJavaObject(clazz);
	}

	/**
	 * 获取整个HashMap
	 * 
	 * @param key 键
	 * @return HashMap
	 */
	public Map getHashMap(String key) {
		return key == null ? null : redisTemplate.opsForHash().entries(key);
	}

	/**
	 * 获取整个HashMap
	 * 
	 * @param key 键
	 * @return HashMap
	 */
	public <T> Map<String, T> getHashMap(String key, Class<T> clazz) {
		Map map = getHashMap(key);
		if (map != null && !map.isEmpty()) {
			Map<String, T> hashMap = new HashMap<String, T>();
			for (Object k : map.keySet()) {
				JSONObject jsonObject = (JSONObject) map.get(k);
				hashMap.put(String.valueOf(k), jsonObject.toJavaObject(clazz));
			}
			return hashMap;
		} else {
			return null;
		}
	}

	/**
	 * 保存整个HashMap
	 * 
	 * @param key 键
	 * @param map 整个HashMap
	 * @return true 成功 false 失败
	 */
	public boolean putHashMap(String key, Map map) {
		try {
			redisTemplate.opsForHash().putAll(key, map);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 保存整个HashMap并设置有效期
	 * 
	 * @param key  键
	 * @param map  整个HashMap
	 * @param time 时间(秒)
	 * @return true成功 false失败
	 */
	public boolean putHashMap(String key, Map map, long time) {
		try {
			redisTemplate.opsForHash().putAll(key, map);
			if (time > 0) {
				expire(key, time);
			}
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 向HashMap中放入数据,如果不存在map将创建
	 * 
	 * @param key   键
	 * @param item  项
	 * @param value 值
	 * @return true 成功 false失败
	 */
	public boolean putHashValue(String key, String item, Object value) {
		try {
			redisTemplate.opsForHash().put(key, item, value);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 *  向HashMap中放入数据并设置有效期,如果不存在map将创建
	 * 
	 * @param key   键
	 * @param item  项
	 * @param value 值
	 * @param time  时间(秒) 注意:如果已存在的hash表有时间,这里将会替换原有的时间
	 * @return true 成功 false失败
	 */
	public boolean putHashValue(String key, String item, Object value, long time) {
		try {
			redisTemplate.opsForHash().put(key, item, value);
			if (time > 0) {
				expire(key, time);
			}
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 删除HashMap中的值
	 * 
	 * @param key  键 不能为null
	 * @param item 项 可以使多个 不能为null
	 */
	public void delHashItem(String key, Object... item) {
		redisTemplate.opsForHash().delete(key, item);
	}

	/**
	 * 判断HashMap中是否有该项的值
	 * 
	 * @param key  键 不能为null
	 * @param item 项 不能为null
	 * @return true 存在 false不存在
	 */
	public boolean hasHashItem(String key, String item) {
		return redisTemplate.opsForHash().hasKey(key, item);
	}

	/**
	 * hash递增 如果不存在,就会创建一个 并把新增后的值返回
	 * 
	 * @param key  键
	 * @param item 项
	 * @param by   要增加几(大于0)
	 * @return
	 */
	public double incrHash(String key, String item, double by) {
		return redisTemplate.opsForHash().increment(key, item, by);
	}

	/**
	 * hash递减
	 * 
	 * @param key  键
	 * @param item 项
	 * @param by   要减少记(小于0)
	 * @return
	 */
	public double decrHash(String key, String item, double by) {
		return redisTemplate.opsForHash().increment(key, item, -by);
	}

	// ===============================set=================================
	/**
	 * 获取整个Set
	 * 
	 * @param key 键
	 * @return
	 */
	public Set getSet(String key) {
		try {
			return redisTemplate.opsForSet().members(key);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 查询Set中是否存在指定的value
	 * 
	 * @param key   键
	 * @param value 值
	 * @return true 存在 false不存在
	 */
	public boolean hasSetValue(String key, Object value) {
		try {
			return redisTemplate.opsForSet().isMember(key, value);
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 弹出Set中的一个值
	 * 
	 * @param key   键
	 * @param value 值
	 * @return true 存在 false不存在
	 */
	public Object popSetValue(String key) {
		try {
			return redisTemplate.opsForSet().pop(key);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 弹出Set中的一个字符串
	 * 
	 * @param key   键
	 * @param value 值
	 * @return true 存在 false不存在
	 */
	public String popSetString(String key) {
		try {
			return (String)redisTemplate.opsForSet().pop(key);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 弹出Set中的一个对象
	 * 
	 * @param key   键
	 * @param value 值
	 * @return true 存在 false不存在
	 */
	public <T> T popSetBean(String key, Class<T> clazz) {
		try {
			JSONObject obj = (JSONObject) redisTemplate.opsForSet().pop(key);
			// 反序列化
			return obj == null ? null : obj.toJavaObject(clazz);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 将数据放入set缓存
	 * 
	 * @param key    键
	 * @param values 值 可以是多个
	 * @return 成功个数
	 */
	public long addSetValue(String key, Object... values) {
		try {
			return redisTemplate.opsForSet().add(key, values);
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

	/**
	 * 将set数据放入缓存，并设置有效期
	 * 
	 * @param key    键
	 * @param time   时间(秒)
	 * @param values 值 可以是多个
	 * @return 成功个数
	 */
	public long addSetValue(String key, long time, Object... values) {
		try {
			Long count = redisTemplate.opsForSet().add(key, values);
			if (time > 0)
				expire(key, time);
			return count;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

	/**
	 * 获取set缓存的长度
	 * 
	 * @param key 键
	 * @return
	 */
	public long getSetSize(String key) {
		try {
			return redisTemplate.opsForSet().size(key);
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

	/**
	 * 移除值为value的
	 * 
	 * @param key    键
	 * @param values 值 可以是多个
	 * @return 移除的个数
	 */
	public long delSetValue(String key, Object... values) {
		try {
			Long count = redisTemplate.opsForSet().remove(key, values);
			return count;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

	// ===============================list=================================
	/**
	 * 获取List的全部值
	 * 
	 * @param key   键
	 * @return
	 */
	public List getListAll(String key) {
		return getList(key, 0, -1);
	}

	/**
	 * 获取List的全部值
	 * 
	 * @param key   键
	 * @return
	 */
	public <T> List<T> getListAll(String key, Class<T> clazz) {
		return getList(key, 0, -1, clazz);
	}

	/**
	 * 获取List中一定的范围
	 * 
	 * @param key   键
	 * @param start 开始
	 * @param end   结束 0 到 -1代表所有值
	 * @return
	 */
	public List getList(String key, long start, long end) {
		try {
			return redisTemplate.opsForList().range(key, start, end);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 获取List中一定的范围
	 * 
	 * @param key   键
	 * @param start 开始
	 * @param end   结束 0 到 -1代表所有值
	 * @return
	 */
	public <T> List<T> getList(String key, long start, long end, Class<T> clazz) {
		try {
			List<T> list = new ArrayList<T>();
			for (Object obj : redisTemplate.opsForList().range(key, start, end)) {
				list.add(((JSONObject)obj).toJavaObject(clazz));
			}
			return list;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 获取list缓存的长度
	 * 
	 * @param key 键
	 * @return
	 */
	public long getListSize(String key) {
		try {
			return redisTemplate.opsForList().size(key);
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

	/**
	 * 通过索引 获取list中的值
	 * 
	 * @param key   键
	 * @param index 索引 index>=0时， 0 表头，1 第二个元素，依次类推；index<0时，-1，表尾，-2倒数第二个元素，依次类推
	 * @return
	 */
	public Object getListValue(String key, long index) {
		try {
			return redisTemplate.opsForList().index(key, index);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 通过索引 获取list中的字符串
	 * 
	 * @param key   键
	 * @param index 索引 index>=0时， 0 表头，1 第二个元素，依次类推；index<0时，-1，表尾，-2倒数第二个元素，依次类推
	 * @return
	 */
	public String getListString(String key, long index) {
		try {
			return (String)redisTemplate.opsForList().index(key, index);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 通过索引 获取list中的对象
	 * 
	 * @param key   键
	 * @param index 索引 index>=0时， 0 表头，1 第二个元素，依次类推；index<0时，-1，表尾，-2倒数第二个元素，依次类推
	 * @return
	 */
	public <T> T getListBean(String key, long index, Class<T> clazz) {
		try {
			JSONObject obj = (JSONObject) redisTemplate.opsForList().index(key, index);
			// 反序列化
			return obj == null ? null : obj.toJavaObject(clazz);
			
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	@Deprecated
	public boolean pushListValue(String key, Object value) {
		return rightPushValue(key, value);
	}

	@Deprecated
	public boolean pushListValue(String key, Object value, long time) {
		return rightPushValue(key, value, time);
	}

	/**
	 * 将值放入list队尾
	 * 
	 * @param key   键
	 * @param value 值
	 * @return
	 */
	public boolean rightPushValue(String key, Object value) {
		try {
			redisTemplate.opsForList().rightPush(key, value);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 将值放入list队尾，并设置有效期
	 * 
	 * @param key   键
	 * @param value 值
	 * @param time  时间(秒)
	 * @return
	 */
	public boolean rightPushValue(String key, Object value, long time) {
		try {
			redisTemplate.opsForList().rightPush(key, value);
			if (time > 0)
				expire(key, time);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 将值放入list队首
	 * 
	 * @param key   键
	 * @param value 值
	 * @return
	 */
	public boolean leftPushValue(String key, Object value) {
		try {
			redisTemplate.opsForList().leftPush(key, value);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 将值放入list队首，并设置有效期
	 * 
	 * @param key   键
	 * @param value 值
	 * @param time  时间(秒)
	 * @return
	 */
	public boolean leftPushValue(String key, Object value, long time) {
		try {
			redisTemplate.opsForList().leftPush(key, value);
			if (time > 0)
				expire(key, time);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 将整个List加入list队尾，如果没有则新建list
	 * 
	 * @param key   键
	 * @param value 值
	 * @return
	 */
	public boolean putList(String key, List value) {
		try {
			redisTemplate.opsForList().rightPushAll(key, value);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 将整个List加入list队尾，并设置有效期，如果没有则新建list
	 *
	 * @param key   键
	 * @param value 值
	 * @param time  时间(秒)
	 * @return
	 */
	public boolean putList(String key, List value, long time) {
		try {
			redisTemplate.opsForList().rightPushAll(key, value);
			if (time > 0)
				expire(key, time);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * 根据索引修改list中的某条数据
	 * 
	 * @param key   键
	 * @param index 索引
	 * @param value 值
	 * @return
	 */
	public boolean updateListValue(String key, long index, Object value) {
		try {
			redisTemplate.opsForList().set(key, index, value);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}
	
	/**
	 * 从队首弹出一个值
	 * @param key
	 * @return
	 */
	public Object leftPopValue(String key) {
		try {
			return redisTemplate.opsForList().leftPop(key);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 从队首弹出一个对象
	 * @param key
	 * @return
	 */
	public <T> T leftPopBean(String key, Class<T> clazz) {
		try {
			JSONObject obj = (JSONObject) redisTemplate.opsForList().leftPop(key);
			// 反序列化
			return obj == null ? null : obj.toJavaObject(clazz);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 从队尾弹出一个值
	 * @param key
	 * @return
	 */
	public Object rightPopValue(String key) {
		try {
			return redisTemplate.opsForList().rightPop(key);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 从队尾弹出一个对象
	 * @param key
	 * @return
	 */
	public <T> T rightPopBean(String key, Class<T> clazz) {
		try {
			JSONObject obj = (JSONObject) redisTemplate.opsForList().rightPop(key);
			// 反序列化
			return obj == null ? null : obj.toJavaObject(clazz);
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * 移除1个value值
	 * 
	 * @param key   键
	 * @param value 值
	 * @return 移除的个数
	 */
	public long removeListValue(String key, Object value) {
		return removeListValue(key, 0, value);
	}

	/**
	 * 移除N个value值
	 * 
	 * @param key   键
	 * @param count 移除多少个
	 * @param value 值
	 * @return 移除的个数
	 */
	public long removeListValue(String key, long count, Object value) {
		try {
			Long remove = redisTemplate.opsForList().remove(key, count, value);
			return remove;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
	}

	// ====================== scan ===========================
	/**
	 * 按照正则表达式遍历
	 * 
	 * @param pattern
	 * @param consumer 以key为参数迭代操作
	 */
	public void scan(String pattern, Consumer<byte[]> consumer) {
		redisTemplate.execute((RedisConnection connection) -> {
			try (Cursor<byte[]> cursor = connection
					.scan(ScanOptions.scanOptions().count(Long.MAX_VALUE).match(pattern).build())) {
				cursor.forEachRemaining(consumer);
				cursor.close();
				return null;
			} catch (IOException e) {
				e.printStackTrace();
				throw new RuntimeException(e);
			}
		});
	}
}
