package cn.spatiotemporal.core.utils;

import java.lang.reflect.Array;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.alibaba.fastjson.serializer.SerializerFeature;

/**
 * 使用fastjson进行POJO的快速序列化/反序列化
 * @author marquis
 *
 */
public class FastJsonUtils {
	
	/**
	 * POJO序列化为json
	 * @param pojo
	 * @return
	 */
	public static String object2Json(Object pojo) {
		return JSON.toJSONString(pojo, SerializerFeature.WriteMapNullValue);
	}

	/**
	 * json反序列化为POJO，需要指定类型
	 * @param json
	 * @param clazz
	 * @return
	 */
	public static <T> T json2Object(String json, Class<T> clazz) {
		if (clazz.isArray()) {
			Class type = clazz.getComponentType();
			JSONArray jarray = JSON.parseArray(json);
			return (T) json2Array(jarray, type);
		} else if (clazz.isAssignableFrom(List.class)) {
			JSONArray jarray = JSON.parseArray(json);
			Class type = (Class) ((ParameterizedType)clazz.getGenericSuperclass())
					.getActualTypeArguments()[0];    //得到泛型类型
			return (T) json2List(jarray, type);
		} else if (clazz.isAssignableFrom(Map.class)) {
			JSONObject jobj = JSON.parseObject(json);
			Class keyType = (Class) ((ParameterizedType)clazz.getGenericSuperclass())
					.getActualTypeArguments()[0];    //得到key的泛型类型
			Class valueType = (Class) ((ParameterizedType)clazz.getGenericSuperclass())
					.getActualTypeArguments()[1];    //得到value的泛型类型
			return (T) json2Map(jobj, keyType, valueType);
		} else if (isSimple(clazz)) {
			return JSON.parseObject(json, clazz);
		}
		
		// 一般POJO
		JSONObject jobj = JSON.parseObject(json);
		return json2Bean(jobj, clazz);
	}
	
	/**
	 * json反序列化为单个POJO
	 * @param jobj
	 * @param clazz
	 * @return
	 */
	public static <T> T json2Bean(JSONObject jobj, Class<T> clazz) {
		try {
			return JSON.parseObject(jobj.toJSONString(), clazz);
//			T obj = clazz.newInstance();
//			PropertyDescriptor[] pds = BeanUtils.getPropertyDescriptors(clazz);
//			for (PropertyDescriptor pd : pds) {
//				String fieldName = pd.getName();
//				if (!"class".equals(fieldName) && ! jobj.containsKey(fieldName)) {
//					Class type = pd.getPropertyType();
////					if (type) {
////						
////					}
//				}
//				
//			}
//			return obj;
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * json反序列化为指定类型的数组
	 * @param jarray
	 * @param clazz
	 * @return
	 */
	public static <T> T[] json2Array(JSONArray jarray, Class<T> clazz) {
		T[] array = (T[]) Array.newInstance(clazz, jarray.size());
		jarray.toArray(array);
//		for (int i = 0; i < jarray.size(); i++) {
//			array[i] = obj2Bean(jarray.getJSONObject(i), clazz);
//		}
		return array;
	}
	
	/**
	 * json反序列化为指定泛型的List
	 * @param jarray
	 * @param clazz
	 * @return
	 */
	public static <T> List json2List(JSONArray jarray, Class<T> clazz) {
		return (List) jarray.toJavaList(clazz);
//		List list = (List) new ArrayList<T>();
//		for (int i = 0; i < jarray.size(); i++) {
//			list.add(json2Bean(jarray.getJSONObject(i), clazz));
//		}
//		return list;
	}
	
	public static <K,V> Map json2Map(JSONObject jobj, Class<K> keyClazz, Class<V> valueClazz) {
		Map map = new HashMap<K, V>();
		for (String key : jobj.keySet()) {
			V value = jobj.toJavaObject(valueClazz);
			map.put(key, value);
			//map.put(key, obj2Bean(jobj.getJSONObject(key), valueClazz));
		}
		return map;
	}
	
	/**
	 * 是否是简单类型
	 * @param clazz
	 * @return
	 */
	public static boolean isSimple(Class clazz) {
		return clazz.isPrimitive()
			|| clazz == java.lang.String.class
			|| clazz == java.util.Date.class
			|| clazz == java.lang.Boolean.class
			|| clazz == java.lang.Byte.class
			|| clazz == java.lang.Character.class
			|| clazz == java.lang.Double.class
			|| clazz == java.lang.Float.class
			|| clazz == java.lang.Integer.class
			|| clazz == java.lang.Long.class
			|| clazz == java.lang.Short.class;
	}
	
	/**
	 * 是否是复杂类型
	 * @param clazz
	 * @return
	 */
	public static boolean isComplex(Class clazz) {
		Type type = clazz;
		return type instanceof GenericArrayType
			&& type instanceof ParameterizedType
			&& type instanceof TypeVariable
			&& type instanceof WildcardType;
	}

	/**
	 * 是否是集合类型
	 * @param clazz
	 * @return
	 */
	public static boolean isCollection(Class clazz) {
		return clazz == Collection.class
			|| clazz == java.util.List.class
			|| clazz == java.util.Set.class
			|| clazz == java.util.Map.class
			|| clazz == java.util.SortedSet.class 
			|| clazz == java.util.SortedMap.class;
	}
}
